/*****************************************************************************\
 * Copyright (c) 2004 Mark Aylett <mark@emantic.co.uk>                       *
 *                                                                           *
 * Permission is hereby granted, free of charge, to any person obtaining a   *
 * copy of this software and associated documentation files (the             *
 * "Software"), to deal in the Software without restriction, including       *
 * without limitation the rights to use, copy, modify, merge, publish,       *
 * distribute, sublicense, and/or sell copies of the Software, and to permit *
 * persons to whom the Software is furnished to do so, subject to the        *
 * following conditions:                                                     *
 *                                                                           *
 * The above copyright notice and this permission notice shall be included   *
 * in all copies or substantial portions of the Software.                    *
 *                                                                           *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS   *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF                *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN *
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,  *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR     *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE *
 * USE OR OTHER DEALINGS IN THE SOFTWARE.                                    *
\*****************************************************************************/

static const char rcsid[] = "$Id: mar_meta_c.c,v 1.11 2004/12/14 13:09:23 marayl Exp $";

#include "mar_meta_c.h"

#include "mar_debug_c.h"
#include "mar_defs_c.h"
#include "mar_format_c.h"
#include "mar_info_c.h"
#include "mar_string_c.h"

#include <assert.h>
#include <errno.h>
#include <string.h>

typedef struct {
    size_t ord_, offset_;
} mar_entry_t;

static void
mar_key(mar_key_t key, const mar_byte_t* index_, size_t ord)
{
    mar_strncpy_(key, (const char*)(index_ + MAR_KEY(ord)), sizeof(mar_key_t));
}

static mar_msize_t
mar_msize(const mar_byte_t* index_, size_t ord)
{
    return mar_decode16_(index_ + MAR_MSIZE(ord));
}

static void
mar_setkey(mar_byte_t* index_, const char* key, size_t ord)
{
    mar_strncpy_((char*)(index_ + MAR_KEY(ord)), key, sizeof(mar_key_t));
}

static void
mar_setmsize(mar_byte_t* index_, mar_msize_t msize, size_t ord)
{
    mar_encode16_(index_ + MAR_MSIZE(ord), msize);
}

static void
mar_entrybyord(mar_entry_t* entry, const mar_byte_t* index_, size_t ord)
{
    /**
     * The resulting offset is the sum of all offsets prior to the specificied
     * ordinal.
     */

    size_t i, offset = 0;
    for (i = 0; ord > i; ++i)
        offset += mar_msize(index_, i);

    entry->ord_ = ord;
    entry->offset_ = offset;
}

static void
mar_entrybykey(mar_entry_t* entry, const mar_info_t info, const char* key)
{
    size_t i, offset = 0;
    for (i = 0; info->isize_ > i; ++i) {

        /**
         * The resulting offset, is the sum of all pairs prior to that with a
         * matching key.
         */

        if (0 == strncmp(key, (const char*)(info->index_ + MAR_KEY(i)),
                         sizeof(mar_key_t)))
            goto done;

        offset += mar_msize(info->index_, i);
    }
 done:
    entry->ord_ = i;
    entry->offset_ = offset;
}

static int
mar_resize(mar_seq_t seq, size_t isize,
           mar_info_t info)
{
    mar_byte_t* addr;
    size_t usize;
    assert(info);

    /**
     * The user size value is cached locally for later restoration; after the
     * sequence has been resized.
     */

    usize = info->usize_;

    if (MAR_ISIZE_MAX < isize) {

        MAR_DEBUG_MSG("Failed to resize meta index:"
                      " Index size exceeded\n");
        errno = EINVAL;
        return -1;
    }

    addr = (mar_byte_t*)mar_resizeseq_(seq, MAR_UDATA(isize));
    if (!addr)
        return -1;

    info->isize_ = isize;
    info->index_ = addr + MAR_INDEX;

    mar_encode16_(addr + MAR_ISIZE, (mar_isize_t)isize);
    mar_encode32_(addr + MAR_USIZE(isize), usize);
    return 0;
}

MAR_EXTERN int
mar_removepairs_(mar_seq_t seq, size_t* isize)
{
    struct mar_info info;
    mar_entry_t entry;
    assert(seq && isize);

    if (-1 == mar_info_(seq, *isize, &info))
        return -1;

    mar_entrybyord(&entry, info.index_, info.isize_);

    if (-1 == mar_resize(seq, *isize = 0, &info))
        return -1;

    if (-1 == mar_setregion_(seq, MAR_META(0, info.usize_), entry.offset_))
        return -1;

    if (!mar_resizeseq_(seq, 0))
        return -1;

    return 0;
}

MAR_EXTERN int
mar_setpair_(mar_seq_t seq, size_t* isize,
             const mar_pair_t* pair, size_t* ord)
{
    struct mar_info info;
    size_t current = 0;
    void* addr;
    mar_entry_t entry;
    assert(seq && isize && pair);

    if (MAR_MSIZE_MAX < pair->size_) {

        MAR_DEBUG_MSG("Failed to set meta pair:"
                      " Meta data size exceeded\n");
        errno = EINVAL;
        return -1;
    }

    if (-1 == mar_info_(seq, *isize, &info))
        return -1;

    mar_entrybykey(&entry, &info, pair->key_);
    if (entry.ord_ == info.isize_) {

        if (-1 == mar_resize(seq, entry.ord_ + 1, &info))
            return -1;

        *isize = entry.ord_ + 1;
        mar_setkey(info.index_, pair->key_, entry.ord_);
    }
    else
        current = mar_msize(info.index_, entry.ord_);

    mar_setmsize(info.index_, (mar_msize_t)pair->size_, entry.ord_);

    if (-1 == mar_setregion_(seq, MAR_META(info.isize_, info.usize_)
                             + entry.offset_, current))
        return -1;

    addr = mar_resizeseq_(seq, pair->size_);
    if (!addr)
        return -1;

    if (pair->data_ && pair->size_)
        memcpy(addr, pair->data_, pair->size_);

    if (ord)
        *ord = entry.ord_;

    return 0;
}

MAR_EXTERN int
mar_setmeta_(mar_seq_t seq, size_t* isize,
             size_t ord, const void* data, size_t size)
{
    struct mar_info info;
    size_t current;
    void* addr;
    mar_entry_t entry;
    assert(seq && isize);

    if (MAR_MSIZE_MAX < size) {

        MAR_DEBUG_MSG("Failed to set meta data:"
                      " Meta data size exceeded\n");
        errno = EINVAL;
        return -1;
    }

    if (-1 == mar_info_(seq, *isize, &info))
        return -1;

    mar_entrybyord(&entry, info.index_, ord);
    if (entry.ord_ == info.isize_) {

        MAR_DEBUG_MSG("Failed to set meta data:"
                      " Pair does not exist\n");
        errno = EINVAL;
        return -1;
    }

    current = mar_msize(info.index_, entry.ord_);
    mar_setmsize(info.index_, (mar_msize_t)size, entry.ord_);

    if (-1 == mar_setregion_(seq, MAR_META(info.isize_, info.usize_)
                             + entry.offset_, current))
        return -1;

    addr = mar_resizeseq_(seq, size);
    if (!addr)
        return -1;

    if (data && size)
        memcpy(addr, data, size);

    return 0;
}

MAR_EXTERN int
mar_unsetbykey_(mar_seq_t seq, size_t* isize,
                const char* key, size_t* ord)
{
    struct mar_info info;
    size_t msize;
    mar_entry_t entry;
    assert(seq && isize && key);

    if (-1 == mar_info_(seq, *isize, &info))
        return -1;

    mar_entrybykey(&entry, &info, key);
    if (entry.ord_ == info.isize_) {

        if (ord)
            *ord = info.isize_;
        return MAR_NOMATCH;
    }

    *isize = info.isize_ - 1;
    msize = mar_msize(info.index_, entry.ord_);

    if (entry.ord_ < *isize) {

        /**
         * If not the last meta pair in the index, then shift all following
         * pairs back one position in order to overwrite the pair being
         * removed.
         */

        mar_byte_t* ptr = info.index_ + MAR_PAIR(entry.ord_);
        memmove(ptr, ptr + MAR_PAIR_SIZE, MAR_PAIR(*isize - entry.ord_));
    }

    /**
     * Remove last index entry.
     */

    if (-1 == mar_resize(seq, *isize, &info))
        return -1;

    /**
     * Set view to and remove data corresponding to removed index entry.
     */

    if (-1 == mar_setregion_(seq, MAR_META(info.isize_, info.usize_)
                             + entry.offset_, msize))
        return -1;

    if (!mar_resizeseq_(seq, 0))
        return -1;

    if (ord)
        *ord = entry.ord_;

    return 0;
}

MAR_EXTERN int
mar_unsetbyord_(mar_seq_t seq, size_t* isize,
                size_t ord)
{
    struct mar_info info;
    mar_entry_t entry;
    size_t msize;
    assert(seq && isize);

    if (-1 == mar_info_(seq, *isize, &info))
        return -1;

    mar_entrybyord(&entry, info.index_, ord);
    if (entry.ord_ == info.isize_)
        return MAR_NOMATCH;

    msize = mar_msize(info.index_, entry.ord_);
    *isize = info.isize_ - 1;

    if (entry.ord_ < *isize) {

        /**
         * If not the last meta pair in the index, then shift all following
         * pairs back one position in order to overwrite the pair being
         * removed.
         */

        mar_byte_t* ptr = info.index_ + MAR_PAIR(entry.ord_);
        memmove(ptr, ptr + MAR_PAIR_SIZE, MAR_PAIR(*isize - entry.ord_));
    }

    /**
     * Remove last index entry.
     */
    if (-1 == mar_resize(seq, *isize, &info))
        return -1;

    /**
     * Set view to and remove data corresponding to removed index entry.
     */

    if (-1 == mar_setregion_(seq, MAR_META(info.isize_, info.usize_)
                             + entry.offset_, msize))
        return -1;

    if (!mar_resizeseq_(seq, 0))
        return -1;

    return 0;
}

MAR_EXTERN const void*
mar_metabykey_(mar_seq_t seq, size_t isize,
               const char* key, size_t* size)
{
    struct mar_info info;
    size_t msize;
    mar_entry_t entry;
    assert(seq && key);

    if (-1 == mar_info_(seq, isize, &info))
        return NULL;

    mar_entrybykey(&entry, &info, key);
    if (entry.ord_ == info.isize_) {

        MAR_DEBUG_MSG("Failed to get meta data by key:"
                      " Pair does not exist\n");
        errno = EINVAL;
        return NULL;
    }

    msize = mar_msize(info.index_, entry.ord_);
    if (-1 == mar_setregion_(seq, MAR_META(info.isize_, info.usize_)
                             + entry.offset_, msize))
        return NULL;

    /**
     * Set optional output parameter.
     */

    if (size)
        *size = msize;

    return mar_seqaddr_(seq);
}

MAR_EXTERN const void*
mar_metabyord_(mar_seq_t seq, size_t isize,
               size_t ord, size_t* size)
{
    struct mar_info info;
    size_t msize;
    mar_entry_t entry;
    assert(seq);

    if (-1 == mar_info_(seq, isize, &info))
        return NULL;

    mar_entrybyord(&entry, info.index_, ord);
    if (entry.ord_ == info.isize_) {

        MAR_DEBUG_MSG("Failed to get meta data by ordinal:"
                      " Pair does not exist\n");
        errno = EINVAL;
        return NULL;
    }

    msize = mar_msize(info.index_, entry.ord_);

    if (-1 == mar_setregion_(seq, MAR_META(info.isize_, info.usize_)
                             + entry.offset_, msize))
        return NULL;

    /**
     * Set optional output parameter.
     */

    if (size)
        *size = msize;

    return mar_seqaddr_(seq);
}

MAR_EXTERN int
mar_metapair_(mar_seq_t seq, size_t isize,
              mar_pair_t* pair, size_t ord)
{
    struct mar_info info;
    mar_entry_t entry;
    assert(seq && pair);

    if (-1 == mar_info_(seq, isize, &info))
        return -1;

    if (ord >= info.isize_) {

        /**
         * Return empty pair on detection of out-of-bounds condition.
         */

        pair->key_[0] = '\0';
        pair->data_ = NULL;
        pair->size_ = 0;
        return MAR_NOMATCH;
    }

    mar_entrybyord(&entry, info.index_, ord);

    /**
     * Assign values from meta index section.
     */

    mar_key(pair->key_, info.index_, ord);
    pair->size_ = mar_msize(info.index_, ord);

    if (-1 == mar_setregion_(seq, MAR_META(info.isize_, info.usize_)
                             + entry.offset_, pair->size_))
        return -1;

    if (!(pair->data_ = mar_seqaddr_(seq)))
        return -1;

    return 0;
}

MAR_EXTERN int
mar_tokey_(mar_seq_t seq, size_t isize,
           mar_key_t key, size_t ord)
{
    struct mar_info info;
    assert(seq && key);

    if (-1 == mar_info_(seq, isize, &info))
        return -1;

    if (ord >= info.isize_) {
        key[0] = '\0';
        return MAR_NOMATCH;
    }

    mar_key(key, info.index_, ord);
    return 0;
}

MAR_EXTERN int
mar_toord_(mar_seq_t seq, size_t isize,
           size_t* ord, const char* key)
{
    struct mar_info info;
    mar_entry_t entry;
    assert(seq && ord && key);

    if (-1 == mar_info_(seq, isize, &info))
        return -1;

    mar_entrybykey(&entry, &info, key);
    *ord = entry.ord_;
    return entry.ord_ == info.isize_ ? MAR_NOMATCH : 0;
}
