/*****************************************************************************\
 * Copyright (c) 2004 Mark Aylett <mark@emantic.co.uk>                       *
 *                                                                           *
 * Permission is hereby granted, free of charge, to any person obtaining a   *
 * copy of this software and associated documentation files (the             *
 * "Software"), to deal in the Software without restriction, including       *
 * without limitation the rights to use, copy, modify, merge, publish,       *
 * distribute, sublicense, and/or sell copies of the Software, and to permit *
 * persons to whom the Software is furnished to do so, subject to the        *
 * following conditions:                                                     *
 *                                                                           *
 * The above copyright notice and this permission notice shall be included   *
 * in all copies or substantial portions of the Software.                    *
 *                                                                           *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS   *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF                *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN *
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,  *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR     *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE *
 * USE OR OTHER DEALINGS IN THE SOFTWARE.                                    *
\*****************************************************************************/

static const char rcsid[] = "$Id: mar_mfile_c.c,v 1.5 2004/12/14 13:09:23 marayl Exp $";

#include "mar_mfile_c.h"

#include "mar_debug_c.h"
#include "mar_file_c.h"
#include "mar_format_c.h"
#include "mar_mmap_c.h"

#include <assert.h>
#include <errno.h>
#include <stdlib.h>

struct mar_mfile {
    int fd_, flags_;
    size_t resvd_, size_;
    mar_mmap_t mmap_;
};

static int
mar_toflags(int from)
{
    int flags;
    switch (from & (MAR_RDONLY | MAR_WRONLY | MAR_RDWR)) {
    case MAR_RDONLY:
        flags = MAR_READ;
        break;
    case MAR_WRONLY:
    case MAR_RDWR:
        flags = MAR_READ | MAR_WRITE;
        break;
    default:
        flags = 0;
    }
    return flags;
}

static size_t
mar_reserve(size_t size)
{
    /**
     * Given the size to be reserved, round up to the nearest page size.
     */
    size_t pagesize = mar_pagesize_();
    size_t pages = size / pagesize;
    if (size % pagesize)
        ++pages;
    return pages * pagesize;
}

MAR_EXTERN int
mar_closemfile_(mar_mfile_t mfile)
{
    int err = 0;
    assert(mfile);

    if (mfile->mmap_) {

        if (-1 == mar_freemmap_(mfile->mmap_))
            err = errno;

        if (mfile->resvd_ > mfile->size_) {

            if (-1 == mar_truncatefile_(mfile->fd_,
                                        (off_t)mfile->size_) && !err)
                err = errno;
        }
    }

    if (-1 == mar_closefile_(mfile->fd_) && !err)
        err = errno;

    free(mfile);
    if (!err)
        return 0;

    errno = err;
    return -1;
}

MAR_EXTERN mar_mfile_t
mar_openmfile_(const char* path, int flags, mode_t mode,
               size_t tail)
{
    int fd;
    size_t size;
    mar_mfile_t mfile;
    assert(path);

    /**
     * Opening a file with the append flag set (where that file is to be
     * memory mapped), does not seem to make much sense. Therefore, the append
     * flag if set is stripped off before the remaining flags are passed to the
     * open file function.
     */
    if (-1 == (fd = mar_openfile_(path, flags & ~MAR_APPEND, mode)))
        return NULL;

    if (-1 == mar_filesize_(fd, &size))
        return NULL;

    if (!(mfile = (mar_mfile_t)malloc(sizeof(struct mar_mfile) + tail)))
        goto fail;

    mfile->fd_ = fd;
    mfile->flags_ = mar_toflags(flags);
    mfile->resvd_ = mfile->size_ = size;
    mfile->mmap_ = NULL;
    return mfile;

 fail:
    mar_closefile_(fd);
    return NULL;
}

MAR_EXTERN void*
mar_mapmfile_(mar_mfile_t mfile, size_t size)
{
    assert(mfile);

    /**
     * Check to see if a large enough area is already mapped.
     */
    if (mfile->mmap_ && mfile->mmap_->len_ >= size)
        goto done;

    if (mfile->resvd_ < size) {

        size_t resvd;

        /**
         * If file has not been opened for writing, then the file cannot be
         * extended.
         */

        if (!(mfile->flags_ & MAR_WRITE)) {

            MAR_DEBUG_MSG("Failed to memory map file:"
                          " File not writable\n");
            errno = EINVAL;
            return NULL;
        }

        resvd = mar_reserve(size);
        if (-1 == mar_extendfile_(mfile->fd_, resvd - mfile->size_))
            return NULL;

        mfile->resvd_ = resvd;
    }

    if (mfile->mmap_) {

        if (-1 == mar_remmap_(mfile->mmap_, 0, size)) {

            int err = errno;
            mar_freemmap_(mfile->mmap_);
            errno = err;
            mfile->mmap_ = NULL;
            return NULL;
        }
    }
    else {

        mfile->mmap_ = mar_createmmap_(mfile->fd_, 0, size, mfile->flags_);
        if (!mfile->mmap_)
            return NULL;
    }

 done:
    if (size > mfile->size_)
        mfile->size_ = size;

    return mfile->mmap_->addr_;
}

MAR_EXTERN int
mar_syncmfile_(mar_mfile_t mfile)
{
    assert(mfile);
    if (mfile->mmap_)
        return mar_syncmmap_(mfile->mmap_);

    return mar_syncfile_(mfile->fd_);
}

MAR_EXTERN int
mar_truncatemfile_(mar_mfile_t mfile, size_t size)
{
    assert(mfile);
    if (!(mfile->flags_ & MAR_WRITE)) {

        MAR_DEBUG_MSG("Failed to truncate memory mapped file:"
                      " File not writable\n");
        errno = EINVAL;
        return -1;
    }
    mfile->size_ = size;
    return 0;
}

MAR_EXTERN void*
mar_mfileaddr_(mar_mfile_t mfile)
{
    assert(mfile);
    if (!mfile->mmap_)
        return NULL;

    return mfile->mmap_->addr_;
}

MAR_EXTERN size_t
mar_mfileresvd_(mar_mfile_t mfile)
{
    assert(mfile);
    return mfile->resvd_;
}

MAR_EXTERN size_t
mar_mfilesize_(mar_mfile_t mfile)
{
    assert(mfile);
    return mfile->size_;
}

MAR_EXTERN void*
mar_mfiletail_(mar_mfile_t mfile)
{
    assert(mfile);
    return (mar_byte_t*)mfile + sizeof(struct mar_mfile);
}

